package morechart.chart.plot

import java.awt.Graphics2D

import java.awt.geom. { Point2D, Rectangle2D }

import org.jfree.chart.axis. { AxisSpace, NumberAxis }
import org.jfree.chart.plot. { Plot, PlotRenderingInfo, PlotState, XYPlot }
import org.jfree.data.RangeType

/** Extends [[org.jfree.chart.plot.XYPlot]] with bounded, expandable and squared axis.
 *
 * @author myst3r10n
 */
trait ExtendedXYPlot extends XYPlot {

  protected var _bounded = false

  /** Bounded mode. */
  def bounded = _bounded

  /** Toggle bounded mode.
   *
   * @param b If this bounded mode is enabled.
   */
  def bounded_= (b: Boolean) { _bounded = b }

  protected var _expandable = true

  /** Expandable mode. */
  def expandable = _expandable

  /** Toggle expandable mode.
   *
   * @param b If this expandable mode is enabled.
   */
  def expandable_= (b: Boolean) { _expandable = b }

  protected var _squared = false

  /** Squared mode. */
  def squared = _squared

  /** Toggle squared mode.
   *
   * @param b If this squared mode is enabled.
   */
  def squared_= (b: Boolean) { _squared = b }

  /** @see [[org.jfree.chart.plot.XYPlot]]. */
  override def draw(g2: Graphics2D,
                    area: Rectangle2D,
                    anchor: Point2D,
                    parentState: PlotState,
                    info: PlotRenderingInfo) {

    // XYPlot part.
    if(area.getWidth <= Plot.MINIMUM_WIDTH_TO_DRAW ||
       area.getHeight <= Plot.MINIMUM_HEIGHT_TO_DRAW)
      return

    val clonedArea = area.clone.asInstanceOf[Rectangle2D]
    getInsets.trim(clonedArea)

    val method = classOf[XYPlot].getDeclaredMethod("calculateAxisSpace",
                                                   classOf[Graphics2D],
                                                   classOf[Rectangle2D])
    method.setAccessible(true)
    val dataArea = method.invoke(this, g2, clonedArea).asInstanceOf[AxisSpace].shrink(clonedArea, null)

    getAxisOffset.trim(dataArea)

    if(dataArea.isEmpty)
      return

    if(bounded)

      // Lock X Axis.
      if(getDomainAxis.isInstanceOf[NumberAxis])
        if(getDomainAxis.asInstanceOf[NumberAxis].getRangeType == RangeType.POSITIVE &&
           getDomainAxis.getLowerBound < 0.0)
          getDomainAxis.setRange(org.jfree.data.Range
            .shift(getDomainAxis.getRange,
                   0 - getDomainAxis.getLowerBound, true))
        else if(getDomainAxis.asInstanceOf[NumberAxis].getRangeType == RangeType.NEGATIVE &&
                getDomainAxis.getUpperBound > 0.0)
          getDomainAxis.setRange(org.jfree.data.Range
            .shift(getDomainAxis.getRange,
                   0 - getDomainAxis.getUpperBound, true))

      // Lock Y Axis.
      if(getRangeAxis.isInstanceOf[NumberAxis])
        if(getRangeAxis.asInstanceOf[NumberAxis].getRangeType == RangeType.POSITIVE &&
           getRangeAxis.getLowerBound < 0.0)
          getRangeAxis.setRange(org.jfree.data.Range
            .shift(getRangeAxis.getRange,
                   0 - getRangeAxis.getLowerBound, true))
        else if(getRangeAxis.asInstanceOf[NumberAxis].getRangeType == RangeType.NEGATIVE &&
                getRangeAxis.getUpperBound > 0.0)
          getRangeAxis.setRange(org.jfree.data.Range
            .shift(getRangeAxis.getRange,
                   0 - getRangeAxis.getUpperBound, true))

    if(expandable) {

      // Expandable X Axis.
      if(lastDataArea.width != dataArea.getWidth) {

        if(lastDataArea.width != -1.0 &&
           lastDataArea.domainLength != -1.0)
          getDomainAxis.setRange(
            new org.jfree.data.Range(
              getDomainAxis.getLowerBound,
              getDomainAxis.getLowerBound + lastDataArea.domainLength /
              lastDataArea.width * dataArea.getWidth), false, false)

        lastDataArea.width = dataArea.getWidth
        lastDataArea.domainLength = getDomainAxis.getRange.getLength

      } else if(lastDataArea.domainLength != getDomainAxis.getRange.getLength)
        lastDataArea.domainLength = getDomainAxis.getRange.getLength

      // Expandable Y Axis.
      if(lastDataArea.height != dataArea.getHeight) {

        if(lastDataArea.height != -1.0 &&
           lastDataArea.rangeLength != -1.0)
          getRangeAxis.setRange(
            new org.jfree.data.Range(
              getRangeAxis.getLowerBound,
                getRangeAxis.getLowerBound + lastDataArea.rangeLength /
                lastDataArea.height * dataArea.getHeight), false, false)

        lastDataArea.height = dataArea.getHeight
        lastDataArea.rangeLength = getRangeAxis.getRange.getLength

      } else if(lastDataArea.rangeLength != getRangeAxis.getRange.getLength)
        lastDataArea.rangeLength = getRangeAxis.getRange.getLength

    }

    // Square Y axis to X axis.
    if(squared &&
       getDomainAxis.getRange.getLength /
       dataArea.getWidth *
       dataArea.getHeight != getRangeAxis.getRange.getLength) {

      getDomainAxis.setRange(
        new org.jfree.data.Range(
          getDomainAxis.getLowerBound,
          getDomainAxis.getLowerBound +
          getRangeAxis.getRange.getLength /
          dataArea.getHeight * dataArea.getWidth), false, false)

      lastDataArea.domainLength = getDomainAxis.getRange.getLength

    }

    // XYPlot part.
    super.draw(g2, area, anchor, parentState, info)

  }


  // Last data.
  private object lastDataArea {

    var width: Double = -1.0
    var height: Double = -1.0
    var domainLength: Double = -1.0
    var rangeLength: Double = -1.0

  }
}

